==> linuxconf-1.19.src.tar.gz: linuxconf-1.19/utils/tlcc.cc <==
#include <stdio.h>
#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>
#include <string.h>

#define TLMP_VERSION	"1.0"

enum FEED_TOKEN {
	T_STR,
	T_MARK,
	T_EOF
};

class FEED{
	const char *fname;
	FILE *fin;
	char buf[1000];
	char *s;
	int noline;
	/*~PROTOBEG~ FEED */
public:
	FEED (const char *_fname, FILE *_fin);
	void fill (void);
	char getcar (void);
	const char *getfname (void)const;
	int getnoline (void)const;
	FEED_TOKEN gettoken (char *str, char *arg1, char *arg2);
	/*~PROTOEND~ FEED */
};

PUBLIC FEED::FEED(const char *_fname, FILE *_fin)
{
	fname = _fname;
	fin = _fin;
	noline = 0;
	fill();
}

PUBLIC void FEED::fill()
{
	s = buf;
	buf[0] = '\0';
	fgets (buf,sizeof(buf)-1,fin);
	noline++;
}

PUBLIC int FEED::getnoline() const
{
	return noline;
}
PUBLIC const char *FEED::getfname() const
{
	return fname;
}

PUBLIC char FEED::getcar()
{
	char ret = '\0';
	if (*s == '\0') fill();
	if (*s != '\0'){
		ret = *s++;
	}
	return ret;
}

static char *tlcc_copystr(char *dst, char *s)
{
	char spccar = *s++;
	*dst++ = spccar;
	while (*s != '\0'){
		if (*s == spccar){
			*dst++ = *s++;
			break;
		}else if (*s == '\\'){
			*dst++ = *s++;
			if (*s != '\0') *dst++ = *s++;
		}else{
			*dst++ = *s++;
		}
	}
	*dst = '\0';
	return s;
}

PUBLIC FEED_TOKEN FEED::gettoken (char *str, char *arg1, char *arg2)
{
	arg1[0] = arg2[0] = str[0] = '\0';
	FEED_TOKEN ret = T_EOF;
	if (*s == '\0') fill();
	char *start = str;
	while (*s != '\0'){
		while (*s != '\0' && *s != '<' && *s != '"' && *s != '\'') *str++ = *s++;
		*str = '\0';
		if (str > start){
			ret = T_STR;
			break;
		}else if (*s == '"' || *s == '\''){
			s = tlcc_copystr (str,s);
			ret = T_STR;
			break;
		}else if (*s == '<'){
			if (isalpha(s[1])){
				char id1[1000],id2[1000],id3[1000];
				char *pt = id1;
				char *ss = s+1;
				while (isalnum(*ss) || *s == '_'){
					*pt++ = *ss++;
				}
				*pt = '\0';
				while (isspace (*ss)) ss++;
				if (*ss == '>'){
					strcpy (str,id1);
					ret = T_MARK;
					s = ss+1;
					break;
				}else if (isalpha(*ss)){
					pt = id2;
					while (isalnum(*ss) || *ss == '_'){
						*pt++ = *ss++;
					}
					*pt = '\0';
					while (isspace (*ss)) ss++;
					if (*ss == '>'){
						ret = T_MARK;
						strcpy (str,id1);
						strcpy (arg1,id2);
						s = ss+1;
						break;
					}else if (isalpha(*ss)){
						pt = id3;
						while (isalnum(*ss) || *ss == '_'){
							*pt++ = *ss++;
						}
						*pt = '\0';
						while (isspace (*ss)) ss++;
						if (*ss == '>'){
							ret = T_MARK;
							strcpy (str,id1);
							strcpy (arg1,id2);
							strcpy (arg2,id3);
							s = ss+1;
							break;
						}
					}
				}else{
					while (s < ss) *str++ = *s++;
				}
			}else if (s[1] == '/'){
				*str++ = '/';
				s += 2;
				while (isalpha(*s)) *str++ = *s++;
				*str = '\0';
				while (isspace (*s)) s++;
				if (*s == '>') s++;
				ret = T_MARK;
				break;
			}else{
				*str++ = *s++;
				if (*s == '\0'){
					*str = '\0';
					ret = T_STR;
					break;
				}
			}
		}
	}
	return ret;
}


class BUFSTR{
	char *buf;
	int size;
	int pos;
	/*~PROTOBEG~ BUFSTR */
public:
	BUFSTR (void);
	void append (char car);
	void append (const char *s);
	void appendf (const char *ctl, ...);
	const char *getbuf (void);
	void reset (void);
	void setnoline (FEED&feed);
	~BUFSTR (void);
	/*~PROTOEND~ BUFSTR */
};

PUBLIC BUFSTR::BUFSTR()
{
	buf = (char*)malloc(1000);
	assert (buf != NULL);
	size = 1000;
	pos = 0;
	buf[0] = '\0';
}

PUBLIC BUFSTR::~BUFSTR()
{
	free (buf);
}

PUBLIC void BUFSTR::append (const char *s)
{
	int len = strlen (s);
	if (pos + len + 1 > size-2){
		size += 1000;
		if (len >= 500) size += len;
		buf = (char*)realloc (buf,size);
		assert (buf != NULL);
	}
	strcpy (buf+pos,s);
	pos += len;
}

PUBLIC void BUFSTR::reset ()
{
	buf[0] = '\0';
	pos = 0;
}

PUBLIC void BUFSTR::setnoline (FEED &feed)
{
	appendf ("\n#line %d \"%s\"\n",feed.getnoline(),feed.getfname());
}

PUBLIC void BUFSTR::appendf (const char *ctl, ...)
{
	va_list list;
	va_start (list,ctl);
	char tmp[1000];
	vsnprintf (tmp,sizeof(tmp)-1,ctl,list);
	va_end (list);
	append (tmp);
}

PUBLIC void BUFSTR::append (char car)
{
	char tmp[2];
	tmp[0] = car;
	tmp[1] = '\0';
	append (tmp);
}

PUBLIC const char *BUFSTR::getbuf()
{
	return buf;
}

class MODSTATE{
public:
	int nomod;
	int nocall;
	bool in_module;
	MODSTATE(){
		nomod = 0;
		nocall = 0;
		in_module = false;
	}
};

static void tlcc_getendparm (FEED &feed, BUFSTR &cur)
{
	int level = 1;
	bool comma = false;
	while (1){
		char carac = feed.getcar();
		if (carac == ')'){
			cur.append (')');
			level--;
			if (level == 0) break;
		}else{
			if (!isspace(carac)){
				if (!comma){
					cur.append (',');
					comma = true;
				}
			}
			cur.append (carac);
		}
	}
}

static void tlcc_getparm (FEED &feed, BUFSTR &cur)
{
	char carac;
	while ((carac = feed.getcar()) <= ' ') cur.append (carac);
	if (carac == '('){
		tlcc_getendparm(feed,cur);
	}
}

static char scope_prefix[100];
static void tlcc_setscopeprefix (const char *fname)
{
	char cwd[PATH_MAX];
	if (getcwd(cwd,sizeof(cwd)-1)!=NULL){
		char *pt = strrchr(cwd,'/');
		if (pt != NULL){
			pt++;
		}else{
			pt = cwd;
		}
		snprintf(scope_prefix,sizeof(scope_prefix)-1,"__sc_%s_%s"
			,pt,fname);
		pt = strchr(scope_prefix,'.');
		pt = scope_prefix;
		while (*pt != '\0'){
			if (!isalpha(*pt) && !isdigit(*pt)) *pt = '_';
			pt++;
		}
		if (pt != NULL) *pt = '\0';
	}else{
		fprintf (stderr,"Can't get the current working directory\n");
		exit (-1);
	}
}

// Put some code to insure that the module exists or that
// the header is include. If not, strange C++ errors are printed
static void tlcc_checkifdef(
	BUFSTR &insert,
	const char *module,
	const char *type)		// "class" or "module"
{
	insert.appendf ("#ifndef _TLMP_%s\n",module);
	insert.appendf ("\t#error TLMP %s %s unknown\n",type,module);
	insert.appendf ("#endif\n");
}

static int tlcc_do (
	FEED &feed,
	BUFSTR &insert,
	BUFSTR &funcdef,
	BUFSTR &mainbuf,
	MODSTATE &st,
	FILE *fout,
	bool glocal_seen)
{
	int ret = 0;
	BUFSTR glocal;			// Definition of all glocals
	BUFSTR callmod;
	BUFSTR callobj;			// Definition and initialisation of the
							// scope object passed to module
							// It is generated right after the glocal
							// instance
	BUFSTR funccore;
	BUFSTR callf;
	BUFSTR *cur = &mainbuf;
	char module[100];
	char s1[1000],s2[1000],s3[1000];
	FEED_TOKEN token;
	int nocall = 0;

	while ((token=feed.gettoken(s1,s2,s3))!=T_EOF){
		//printf ("gettoken %d :%s: :%s:\n",token,s1,s2);
		if (token == T_STR){
			cur->append (s1);
		}else if (strcmp(s1,"glocal")==0){
			glocal_seen = true;
			st.nomod++;
			insert.appendf ("struct __GLOCAL%d{\n",st.nomod);
			cur = &insert;
		}else if (strcmp(s1,"/glocal")==0){
			insert.append ("};\n");
			funcdef.appendf ("\t__GLOCAL%d glocal;\n",st.nomod);
			cur = &funccore;
			funccore.setnoline (feed);
		}else if (strcmp (s1,"mod")==0){
			insert.append (glocal.getbuf());
			insert.append (callobj.getbuf());
			insert.append (callf.getbuf());
			mainbuf.append (funccore.getbuf());
			fputs (insert.getbuf(),fout);
			fputs (funcdef.getbuf(),fout);
			fputs (mainbuf.getbuf(),fout);
			insert.reset();
			mainbuf.reset();
			glocal.reset();
			funcdef.reset();
			funccore.reset();
			callobj.reset();
			callf.reset();

			cur = &funcdef;
			funcdef.setnoline (feed);
			st.in_module = true;
		}else if (strcmp (s1,"/mod")==0){
			glocal_seen = false;
			st.in_module = false;
		}else if (strcmp (s1,"call")==0 ){
			st.nocall++;
			nocall = st.nocall;
			if (glocal_seen){
				funcdef.appendf ("\t%s%d _scopeobj%d(glocal);\n"
					,scope_prefix,nocall,nocall);
			}else{
				funcdef.appendf ("\t%s%d _scopeobj%d;\n"
					,scope_prefix,nocall,nocall);
			}
			callmod.reset();
			strcpy (module,s2);
			tlcc_checkifdef(insert,module,"module");
			cur->setnoline(feed);
			cur->append (module);
			cur->appendf ("(_scopeobj%d",nocall);
			tlcc_getparm (feed,*cur);
			callobj.appendf ("class %s%d: public _F_%s{\n"
					,scope_prefix,nocall,module);
			callobj.append ("public:\n");
			if (glocal_seen){
				callobj.appendf ("\t__GLOCAL%d &glocal;\n",st.nomod);
				callobj.appendf ("\t%s%d (__GLOCAL%d &g)\n"
					,scope_prefix,nocall,st.nomod);
				callobj.append ("\t\t: glocal(g)\n");
			}else{
				callobj.appendf ("\t%s%d ()\n",scope_prefix,nocall);
			}
			callobj.append ("\t{\n\t}\n");
		}else if (strcmp (s1,"obj")==0 ){
			if (!st.in_module){
				insert.append (glocal.getbuf());
				insert.append (callobj.getbuf());
				insert.append (callf.getbuf());
				mainbuf.append (funccore.getbuf());
				fputs (insert.getbuf(),fout);
				fputs (funcdef.getbuf(),fout);
				fputs (mainbuf.getbuf(),fout);
				insert.reset();
				mainbuf.reset();
				glocal.reset();
				funcdef.reset();
				funccore.reset();
				callobj.reset();
				callf.reset();
			}
			st.nocall++;
			nocall = st.nocall;
			if (glocal_seen){
				funcdef.appendf ("\t%s%d _scopeobj%d(glocal);\n"
					,scope_prefix,nocall,nocall);
			}else{
				funcdef.appendf ("\t%s%d _scopeobj%d;\n"
					,scope_prefix,nocall,nocall);
			}
			callmod.reset();
			strcpy (module,s2);
			tlcc_checkifdef(insert,module,"class");
			cur->setnoline(feed);
			cur->appendf ("%s %s",module,s3);
			{
				cur->appendf ("(_scopeobj%d",nocall);
				char carac;
				while (isspace(carac = feed.getcar()));
				if (carac == ';'){
					cur->append (");");
				}else if (carac == '('){
					tlcc_getendparm (feed,*cur);
				}else{
					fprintf (stderr,"Invalid obj def line %d\n",feed.getnoline());
				}
			}
			callobj.appendf ("class %s%d: public _F_%s{\n"
					,scope_prefix,nocall,module);
			callobj.append ("public:\n");
			if (glocal_seen){
				callobj.appendf ("\t__GLOCAL%d &glocal;\n",st.nomod);
				callobj.appendf ("\t%s%d (__GLOCAL%d &g)\n"
					,scope_prefix,nocall,st.nomod);
				callobj.append ("\t\t: glocal(g)\n");
			}else{
				callobj.appendf ("\t%s%d ()\n",scope_prefix,nocall);
			}
			callobj.append ("\t{\n\t}\n");
			#if 0
			if (glocal_seen){
				funcdef.appendf ("\t%s%d _scopeobj%d(glocal);\n"
					,scope_prefix,nocall,nocall);
			}else{
				funcdef.appendf ("\t%s%d _scopeobj%d();\n"
					,scope_prefix,nocall,nocall);
			}
			#endif
		}else if (strcmp(s1,"/call")==0 || strcmp(s1,"/obj")==0){
			callobj.append ("};\n");
		}else if (strcmp (s1,"f")==0){
			// Put some code to insure that the logfunc exists
			// If not, strange C++ errors are printed
			insert.appendf ("#ifndef _F_%s_%s\n",module,s2);
			insert.appendf ("\t#error Unknown logical function %s for module %s\n"
				,s2,module);
			insert.appendf ("#endif\n");
			//funcdef.appendf ("\t_scopeobj%d.%s=%s_%s__%d;\n"
			//	,nocall,s2,module,s2,nocall);
			callobj.appendf ("\t_F_%s_%s( );\n",module,s2,nocall);
			callf.appendf ("_F_%s_%s(%s%d::)\n",module,s2,scope_prefix,nocall);
			callf.append  ("{\n");
			//callf.appendf  ("\t__GLOCAL%d &glocal = ((%s%d*) this)->glocal;\n"
			//	,st.nomod,scope_prefix,nocall);
			callf.setnoline(feed);
			BUFSTR endcallf;
			tlcc_do (feed,insert,callf,endcallf,st,fout,glocal_seen);
			callf.append (endcallf.getbuf());
		}else if (strcmp (s1,"/f")==0){
			mainbuf.append ("}\n");
			mainbuf.setnoline (feed);
			break;
		}
	}
	insert.append (glocal.getbuf());
	insert.append (callobj.getbuf());
	insert.append (callf.getbuf());
	mainbuf.append (funccore.getbuf());
	return ret;
}

static void usage()
{
	fprintf (stderr
		,"tlmp translator version %s\n"
		 "\n"
		 "tlcc input-file [ output-file ]\n",TLMP_VERSION);
}
int main (int argc, char *argv[])
{
	int ret = 0;
	if (argc  != 3 && argc != 2){
		usage();
		ret = -1;
	}else{
		tlcc_setscopeprefix (argv[1]);
		FILE *fin = fopen (argv[1],"r");
		if (fin == NULL){
			fprintf (stderr,"Can't open file %s (%s)\n",argv[1]
				,strerror(errno));
		}else{
			FILE *fout = stdout;
			if (argc == 3) fout = fopen (argv[2],"w");
			if (fout == NULL){
				fprintf (stderr,"Can't open file %s (%s)\n",argv[2]
					,strerror(errno));
			}else{
				FEED feed (argv[1],fin);
				MODSTATE state;
				BUFSTR insert,funcdef,mainbuf;
				tlcc_do (feed,insert,funcdef,mainbuf,state,fout,false);
				fputs (insert.getbuf(),fout);
				fputs (funcdef.getbuf(),fout);
				fputs (mainbuf.getbuf(),fout);
				if (fout != stdout && fclose (fout) != 0) ret = -1;
			}
			fclose (fin);
		}
	}
	return ret;
}

==> linuxconf-1.20.src.tar.gz: linuxconf-1.20/utils/tlcc.cc <==
#include <stdio.h>
#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>
#include <string.h>

#define TLMP_VERSION	"1.0"

enum FEED_TOKEN {
	T_STR,
	T_MARK,
	T_EOF
};

class FEED{
	const char *fname;
	FILE *fin;
	char buf[1000];
	char *s;
	int noline;
	/*~PROTOBEG~ FEED */
public:
	FEED (const char *_fname, FILE *_fin);
	void fill (void);
	char getcar (void);
	const char *getfname (void)const;
	int getnoline (void)const;
	FEED_TOKEN gettoken (char *str, char *arg1, char *arg2);
	/*~PROTOEND~ FEED */
};

PUBLIC FEED::FEED(const char *_fname, FILE *_fin)
{
	fname = _fname;
	fin = _fin;
	noline = 0;
	fill();
}

PUBLIC void FEED::fill()
{
	s = buf;
	buf[0] = '\0';
	fgets (buf,sizeof(buf)-1,fin);
	noline++;
}

PUBLIC int FEED::getnoline() const
{
	return noline;
}
PUBLIC const char *FEED::getfname() const
{
	return fname;
}

PUBLIC char FEED::getcar()
{
	char ret = '\0';
	if (*s == '\0') fill();
	if (*s != '\0'){
		ret = *s++;
	}
	return ret;
}

static char *tlcc_copystr(char *dst, char *s)
{
	char spccar = *s++;
	*dst++ = spccar;
	while (*s != '\0'){
		if (*s == spccar){
			*dst++ = *s++;
			break;
		}else if (*s == '\\'){
			*dst++ = *s++;
			if (*s != '\0') *dst++ = *s++;
		}else{
			*dst++ = *s++;
		}
	}
	*dst = '\0';
	return s;
}

PUBLIC FEED_TOKEN FEED::gettoken (char *str, char *arg1, char *arg2)
{
	arg1[0] = arg2[0] = str[0] = '\0';
	FEED_TOKEN ret = T_EOF;
	if (*s == '\0') fill();
	char *start = str;
	while (*s != '\0'){
		while (*s != '\0' && *s != '<' && *s != '"' && *s != '\'') *str++ = *s++;
		*str = '\0';
		if (str > start){
			ret = T_STR;
			break;
		}else if (*s == '"' || *s == '\''){
			s = tlcc_copystr (str,s);
			ret = T_STR;
			break;
		}else if (*s == '<'){
			if (isalpha(s[1])){
				char id1[1000],id2[1000],id3[1000];
				char *pt = id1;
				char *ss = s+1;
				while (isalnum(*ss) || *s == '_'){
					*pt++ = *ss++;
				}
				*pt = '\0';
				while (isspace (*ss)) ss++;
				if (*ss == '>'){
					strcpy (str,id1);
					ret = T_MARK;
					s = ss+1;
					break;
				}else if (isalpha(*ss)){
					pt = id2;
					while (isalnum(*ss) || *ss == '_'){
						*pt++ = *ss++;
					}
					*pt = '\0';
					while (isspace (*ss)) ss++;
					if (*ss == '>'){
						ret = T_MARK;
						strcpy (str,id1);
						strcpy (arg1,id2);
						s = ss+1;
						break;
					}else if (isalpha(*ss)){
						pt = id3;
						while (isalnum(*ss) || *ss == '_'){
							*pt++ = *ss++;
						}
						*pt = '\0';
						while (isspace (*ss)) ss++;
						if (*ss == '>'){
							ret = T_MARK;
							strcpy (str,id1);
							strcpy (arg1,id2);
							strcpy (arg2,id3);
							s = ss+1;
							break;
						}
					}
				}else{
					while (s < ss) *str++ = *s++;
				}
			}else if (s[1] == '/'){
				*str++ = '/';
				s += 2;
				while (isalpha(*s)) *str++ = *s++;
				*str = '\0';
				while (isspace (*s)) s++;
				if (*s == '>') s++;
				ret = T_MARK;
				break;
			}else{
				*str++ = *s++;
				if (*s == '\0'){
					*str = '\0';
					ret = T_STR;
					break;
				}
			}
		}
	}
	return ret;
}


class BUFSTR{
	char *buf;
	int size;
	int pos;
	/*~PROTOBEG~ BUFSTR */
public:
	BUFSTR (void);
	void append (char car);
	void append (const char *s);
	void appendf (const char *ctl, ...);
	const char *getbuf (void);
	void reset (void);
	void setnoline (FEED&feed);
	~BUFSTR (void);
	/*~PROTOEND~ BUFSTR */
};

PUBLIC BUFSTR::BUFSTR()
{
	buf = (char*)malloc(1000);
	assert (buf != NULL);
	size = 1000;
	pos = 0;
	buf[0] = '\0';
}

PUBLIC BUFSTR::~BUFSTR()
{
	free (buf);
}

PUBLIC void BUFSTR::append (const char *s)
{
	int len = strlen (s);
	if (pos + len + 1 > size-2){
		size += 1000;
		if (len >= 500) size += len;
		buf = (char*)realloc (buf,size);
		assert (buf != NULL);
	}
	strcpy (buf+pos,s);
	pos += len;
}

PUBLIC void BUFSTR::reset ()
{
	buf[0] = '\0';
	pos = 0;
}

PUBLIC void BUFSTR::setnoline (FEED &feed)
{
	appendf ("\n#line %d \"%s\"\n",feed.getnoline(),feed.getfname());
}

PUBLIC void BUFSTR::appendf (const char *ctl, ...)
{
	va_list list;
	va_start (list,ctl);
	char tmp[1000];
	vsnprintf (tmp,sizeof(tmp)-1,ctl,list);
	va_end (list);
	append (tmp);
}

PUBLIC void BUFSTR::append (char car)
{
	char tmp[2];
	tmp[0] = car;
	tmp[1] = '\0';
	append (tmp);
}

PUBLIC const char *BUFSTR::getbuf()
{
	return buf;
}

class MODSTATE{
public:
	int nomod;
	int nocall;
	bool in_module;
	MODSTATE(){
		nomod = 0;
		nocall = 0;
		in_module = false;
	}
};

static void tlcc_getendparm (FEED &feed, BUFSTR &cur)
{
	int level = 1;
	bool comma = false;
	while (1){
		char carac = feed.getcar();
		if (carac == ')'){
			cur.append (')');
			level--;
			if (level == 0) break;
		}else{
			if (!isspace(carac)){
				if (!comma){
					cur.append (',');
					comma = true;
				}
			}
			cur.append (carac);
		}
	}
}

static void tlcc_getparm (FEED &feed, BUFSTR &cur)
{
	char carac;
	while ((carac = feed.getcar()) <= ' ') cur.append (carac);
	if (carac == '('){
		tlcc_getendparm(feed,cur);
	}
}

static char scope_prefix[100];
static void tlcc_setscopeprefix (const char *fname)
{
	char cwd[PATH_MAX];
	if (getcwd(cwd,sizeof(cwd)-1)!=NULL){
		char *pt = strrchr(cwd,'/');
		if (pt != NULL){
			pt++;
		}else{
			pt = cwd;
		}
		snprintf(scope_prefix,sizeof(scope_prefix)-1,"__sc_%s_%s"
			,pt,fname);
		pt = strchr(scope_prefix,'.');
		pt = scope_prefix;
		while (*pt != '\0'){
			if (!isalpha(*pt) && !isdigit(*pt)) *pt = '_';
			pt++;
		}
		if (pt != NULL) *pt = '\0';
	}else{
		fprintf (stderr,"Can't get the current working directory\n");
		exit (-1);
	}
}

// Put some code to insure that the module exists or that
// the header is include. If not, strange C++ errors are printed
static void tlcc_checkifdef(
	BUFSTR &insert,
	const char *module,
	const char *type)		// "class" or "module"
{
	insert.appendf ("#ifndef _TLMP_%s\n",module);
	insert.appendf ("\t#error TLMP %s %s unknown\n",type,module);
	insert.appendf ("#endif\n");
}

static int tlcc_do (
	FEED &feed,
	BUFSTR &insert,
	BUFSTR &funcdef,
	BUFSTR &mainbuf,
	MODSTATE &st,
	FILE *fout,
	bool glocal_seen)
{
	int ret = 0;
	BUFSTR glocal;			// Definition of all glocals
	BUFSTR callmod;
	BUFSTR callobj;			// Definition and initialisation of the
							// scope object passed to module
							// It is generated right after the glocal
							// instance
	BUFSTR funccore;
	BUFSTR callf;
	BUFSTR *cur = &mainbuf;
	char module[100];
	char s1[1000],s2[1000],s3[1000];
	FEED_TOKEN token;
	int nocall = 0;

	while ((token=feed.gettoken(s1,s2,s3))!=T_EOF){
		//printf ("gettoken %d :%s: :%s:\n",token,s1,s2);
		if (token == T_STR){
			cur->append (s1);
		}else if (strcmp(s1,"glocal")==0){
			glocal_seen = true;
			st.nomod++;
			insert.appendf ("struct __GLOCAL%d{\n",st.nomod);
			cur = &insert;
		}else if (strcmp(s1,"/glocal")==0){
			insert.append ("};\n");
			funcdef.appendf ("\t__GLOCAL%d glocal;\n",st.nomod);
			cur = &funccore;
			funccore.setnoline (feed);
		}else if (strcmp (s1,"mod")==0){
			insert.append (glocal.getbuf());
			insert.append (callobj.getbuf());
			insert.append (callf.getbuf());
			mainbuf.append (funccore.getbuf());
			fputs (insert.getbuf(),fout);
			fputs (funcdef.getbuf(),fout);
			fputs (mainbuf.getbuf(),fout);
			insert.reset();
			mainbuf.reset();
			glocal.reset();
			funcdef.reset();
			funccore.reset();
			callobj.reset();
			callf.reset();

			cur = &funcdef;
			funcdef.setnoline (feed);
			st.in_module = true;
		}else if (strcmp (s1,"/mod")==0){
			glocal_seen = false;
			st.in_module = false;
		}else if (strcmp (s1,"call")==0 ){
			st.nocall++;
			nocall = st.nocall;
			if (glocal_seen){
				funcdef.appendf ("\t%s%d _scopeobj%d(glocal);\n"
					,scope_prefix,nocall,nocall);
			}else{
				funcdef.appendf ("\t%s%d _scopeobj%d;\n"
					,scope_prefix,nocall,nocall);
			}
			callmod.reset();
			strcpy (module,s2);
			tlcc_checkifdef(insert,module,"module");
			cur->setnoline(feed);
			cur->append (module);
			cur->appendf ("(_scopeobj%d",nocall);
			tlcc_getparm (feed,*cur);
			callobj.appendf ("class %s%d: public _F_%s{\n"
					,scope_prefix,nocall,module);
			callobj.append ("public:\n");
			if (glocal_seen){
				callobj.appendf ("\t__GLOCAL%d &glocal;\n",st.nomod);
				callobj.appendf ("\t%s%d (__GLOCAL%d &g)\n"
					,scope_prefix,nocall,st.nomod);
				callobj.append ("\t\t: glocal(g)\n");
			}else{
				callobj.appendf ("\t%s%d ()\n",scope_prefix,nocall);
			}
			callobj.append ("\t{\n\t}\n");
		}else if (strcmp (s1,"obj")==0 ){
			if (!st.in_module){
				insert.append (glocal.getbuf());
				insert.append (callobj.getbuf());
				insert.append (callf.getbuf());
				mainbuf.append (funccore.getbuf());
				fputs (insert.getbuf(),fout);
				fputs (funcdef.getbuf(),fout);
				fputs (mainbuf.getbuf(),fout);
				insert.reset();
				mainbuf.reset();
				glocal.reset();
				funcdef.reset();
				funccore.reset();
				callobj.reset();
				callf.reset();
			}
			st.nocall++;
			nocall = st.nocall;
			if (glocal_seen){
				funcdef.appendf ("\t%s%d _scopeobj%d(glocal);\n"
					,scope_prefix,nocall,nocall);
			}else{
				funcdef.appendf ("\t%s%d _scopeobj%d;\n"
					,scope_prefix,nocall,nocall);
			}
			callmod.reset();
			strcpy (module,s2);
			tlcc_checkifdef(insert,module,"class");
			cur->setnoline(feed);
			cur->appendf ("%s %s",module,s3);
			{
				cur->appendf ("(_scopeobj%d",nocall);
				char carac;
				while (isspace(carac = feed.getcar()));
				if (carac == ';'){
					cur->append (");");
				}else if (carac == '('){
					tlcc_getendparm (feed,*cur);
				}else{
					fprintf (stderr,"Invalid obj def line %d\n",feed.getnoline());
				}
			}
			callobj.appendf ("class %s%d: public _F_%s{\n"
					,scope_prefix,nocall,module);
			callobj.append ("public:\n");
			if (glocal_seen){
				callobj.appendf ("\t__GLOCAL%d &glocal;\n",st.nomod);
				callobj.appendf ("\t%s%d (__GLOCAL%d &g)\n"
					,scope_prefix,nocall,st.nomod);
				callobj.append ("\t\t: glocal(g)\n");
			}else{
				callobj.appendf ("\t%s%d ()\n",scope_prefix,nocall);
			}
			callobj.append ("\t{\n\t}\n");
			#if 0
			if (glocal_seen){
				funcdef.appendf ("\t%s%d _scopeobj%d(glocal);\n"
					,scope_prefix,nocall,nocall);
			}else{
				funcdef.appendf ("\t%s%d _scopeobj%d();\n"
					,scope_prefix,nocall,nocall);
			}
			#endif
		}else if (strcmp(s1,"/call")==0 || strcmp(s1,"/obj")==0){
			callobj.append ("};\n");
		}else if (strcmp (s1,"f")==0){
			// Put some code to insure that the logfunc exists
			// If not, strange C++ errors are printed
			insert.appendf ("#ifndef _F_%s_%s\n",module,s2);
			insert.appendf ("\t#error Unknown logical function %s for module/class %s\n"
				,s2,module);
			insert.appendf ("#endif\n");
			//funcdef.appendf ("\t_scopeobj%d.%s=%s_%s__%d;\n"
			//	,nocall,s2,module,s2,nocall);
			callobj.appendf ("\t_F_%s_%s( );\n",module,s2,nocall);
			callf.appendf ("_F_%s_%s(%s%d::)\n",module,s2,scope_prefix,nocall);
			callf.append  ("{\n");
			//callf.appendf  ("\t__GLOCAL%d &glocal = ((%s%d*) this)->glocal;\n"
			//	,st.nomod,scope_prefix,nocall);
			callf.setnoline(feed);
			BUFSTR endcallf;
			tlcc_do (feed,insert,callf,endcallf,st,fout,glocal_seen);
			callf.append (endcallf.getbuf());
		}else if (strcmp (s1,"/f")==0){
			mainbuf.append ("}\n");
			mainbuf.setnoline (feed);
			break;
		}else{
			fprintf (stderr,"Unknown directive %s\n",s1);
			ret = -1;
		}
	}
	insert.append (glocal.getbuf());
	insert.append (callobj.getbuf());
	insert.append (callf.getbuf());
	mainbuf.append (funccore.getbuf());
	return ret;
}

static void usage()
{
	fprintf (stderr
		,"tlmp translator version %s\n"
		 "\n"
		 "tlcc input-file [ output-file ]\n",TLMP_VERSION);
}
int main (int argc, char *argv[])
{
	int ret = 0;
	if (argc  != 3 && argc != 2){
		usage();
		ret = -1;
	}else{
		tlcc_setscopeprefix (argv[1]);
		FILE *fin = fopen (argv[1],"r");
		if (fin == NULL){
			fprintf (stderr,"Can't open file %s (%s)\n",argv[1]
				,strerror(errno));
		}else{
			FILE *fout = stdout;
			if (argc == 3) fout = fopen (argv[2],"w");
			if (fout == NULL){
				fprintf (stderr,"Can't open file %s (%s)\n",argv[2]
					,strerror(errno));
			}else{
				FEED feed (argv[1],fin);
				MODSTATE state;
				BUFSTR insert,funcdef,mainbuf;
				tlcc_do (feed,insert,funcdef,mainbuf,state,fout,false);
				fputs (insert.getbuf(),fout);
				fputs (funcdef.getbuf(),fout);
				fputs (mainbuf.getbuf(),fout);
				if (fout != stdout && fclose (fout) != 0) ret = -1;
			}
			fclose (fin);
		}
	}
	return ret;
}
